from safe_autonomy_sims.gym.inspection.inspection_v0 import InspectionEnv
from safe_autonomy_sims.gym.inspection.weighted_inspection_v0 import WeightedInspectionEnv
from safe_autonomy_sims.gym.inspection.sixdof_inspection_v0 import WeightedSixDofInspectionEnv


__all__ = ['InspectionEnv', 'WeightedInspectionEnv', 'WeightedSixDofInspectionEnv']



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements done functions for safe autonomy reinforcement
learning environments.

Done functions are called by [simulators](../simulators/index.md)
at each timestep of the simulation. These functions are tasked with
determining when an agent has reached a done condition for their
current training episode on a given task. They take in the current
and previous observation and action spaces and output a boolean value.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements agents for the docking and inspection
environments.

Agents are actors which control [platforms](../platforms/index.md)
within an environment. They make decisions based on a learned or
programmed policy given their current observation space. The agent's
actions influence the environment via platform controls.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a Reward function that rewards the agent for ever live
timestep up to some maximum
"""
import logging
import typing

import numpy as np
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator


class LiveTimestepRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for LiveTimestepReward

    Attributes
    ----------
    step_reward : float
        the maximum reward value for each timestep
    max_time_rewarded : float
        the maximum simulation time.  Reward is zero for all
        timesteps for which sim_time > max_time_rewarded [Note use of sim_time
        instead of timestep count!]
    """
    step_reward: float = 0.01
    max_time_rewarded: float = np.inf


class LiveTimestepReward(RewardFuncBase):
    """
    Gives reward for a discrete action matching its target
    """

    @staticmethod
    def get_validator() -> typing.Type[LiveTimestepRewardValidator]:
        return LiveTimestepRewardValidator

    def __init__(self, **kwargs) -> None:
        self.config: LiveTimestepRewardValidator
        super().__init__(**kwargs)
        self._logger = logging.getLogger(self.name)

    def __call__(self, observation, action, next_observation, state, next_state, observation_space, observation_units) -> float:

        reward = 0.0

        if self.config.agent_name not in next_observation:
            return reward

        sim_time = next_state.sim_time

        if sim_time <= self.config.max_time_rewarded:
            reward = self.config.step_reward

        return reward



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains common controllers for the platforms.
"""

import numpy as np
from corl.libraries.plugin_library import PluginLibrary  # pylint: disable=E0401
from corl.libraries.units import corl_get_ureg
from corl.simulators.base_parts import BaseController, BasePlatformPartValidator  # pylint: disable=E0401
from pydantic import BaseModel
from pydantic.types import PyObject

from safe_autonomy_sims.platforms.cwh.cwh_available_platforms import CWHAvailablePlatformTypes
from safe_autonomy_sims.simulators.cwh_simulator import CWHSimulator
from safe_autonomy_sims.simulators.inspection_simulator import InspectionSimulator


class CommonController(BaseController):
    """
    A controller created to interface with the CWH platform.
    """

    def apply_control(self, control: np.ndarray) -> None:
        """
        Raises
        ------
        NotImplementedError
            If the method has not been implemented
        """
        raise NotImplementedError

    def get_applied_control(self) -> np.ndarray:
        """
        Raises
        ------
        NotImplementedError
            If the method has not been implemented
        """
        raise NotImplementedError


class RateControllerValidator(BasePlatformPartValidator):
    """
    A configuration validator for RateController

    Attributes
    ----------
    axis: int
        Index in combined control vector for this controller's output action.
    """
    axis: int


class ControllerPropValidator(BaseModel):
    """
    A configuration validator for ControllerProp

    Attributes
    ----------
    property_class: PyObject
        The Prop class defining the controller's bounds and units
    """
    property_class: PyObject


# TODO: Find a better way to save actions to platform instead of storing an axis in the saved vector
class RateController(CommonController):
    """
    Generic rate controller. Writes control action to platform's control vector and reads applied action from platform.
    """

    def __init__(
        self,
        parent_platform,
        config,
    ):  # pylint: disable=W0102
        self.config: RateControllerValidator
        self.prop_config: ControllerPropValidator = self.get_prop_validator(**config)  # get property class
        # TODO: there must be a better way to get unit info
        self.unit = self.prop_config.property_class().unit
        super().__init__(property_class=self.prop_config.property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator():
        """
        Property to return the generic RateController validator.
        """
        return RateControllerValidator

    @property
    def get_prop_validator(self):
        """
        Property to return the generic RateController validator.
        """
        return ControllerPropValidator

    def apply_control(self, control: np.ndarray) -> None:
        self.parent_platform.save_action_to_platform(action=control, axis=self.config.axis)

    def get_applied_control(self) -> np.ndarray:
        return corl_get_ureg().Quantity(
            np.array([self.parent_platform.get_applied_action().m[self.config.axis]], dtype=np.float32), self.unit
        )


for sim in [CWHSimulator, InspectionSimulator]:
    PluginLibrary.AddClassToGroup(RateController, "RateController", {"simulator": sim, "platform_type": CWHAvailablePlatformTypes})



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements initializers for the safe autonomy simulators.

Initializers are responsible for setting up the initial conditions
of platforms in a simulation. Given a set of sampled parameters,
initializers return a set of initialization parameters expected by
the platforms and their associated simulation entities.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements glues for the six-dof environments
"""
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np
from corl.glues.base_multi_wrapper import BaseMultiWrapperGlue, BaseMultiWrapperGlueValidator
from corl.glues.base_wrapper import BaseWrapperGlue
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg
from pydantic import validator
from scipy.spatial.transform import Rotation


class MagNorm3DGlue(BaseWrapperGlue):
    """Wrapper glue that converts output vector of wrapped glue into the form [magnitude, normalized unit vector]"""

    class Fields:
        """
        Fields in this glue
        """
        DIRECT_OBSERVATION = "direct_observation"

    def get_unique_name(self) -> str:
        return self.glue().get_unique_name() + "_MagNorm3D"

    @cached_property
    def observation_prop(self):
        # unit = self.glue().observation_space[self.glue().Fields.DIRECT_OBSERVATION].unit
        prop = BoxProp(low=[-10000, -1, -1, -1], high=[10000, 1, 1, 1], unit="dimensionless")
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: prop})

    @cached_property
    def observation_space(self) -> gymnasium.spaces.Space:
        wrapped_space = self.glue().observation_space[self.glue().Fields.DIRECT_OBSERVATION]

        mag = np.linalg.norm(np.maximum(np.abs(wrapped_space.low), np.abs(wrapped_space.high)))

        low = np.concatenate([[0], -1 * np.ones(3)], dtype=np.float32)  # pylint: disable=unexpected-keyword-arg
        high = np.concatenate([[mag], np.ones(3)], dtype=np.float32)  # pylint: disable=unexpected-keyword-arg

        d = gymnasium.spaces.dict.Dict()
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(low=low, high=high, dtype=np.float32)
        return d

    @cached_property
    def normalized_observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        return self.observation_space

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        obs = self.glue().get_observation(other_obs, obs_space, obs_units)[self.glue().Fields.DIRECT_OBSERVATION]
        mag = np.linalg.norm(obs.m)

        # if mag == 0:
        if mag < 1e-5:
            output = np.concatenate([[0], np.zeros_like(obs.m)], dtype=np.float32)  # pylint: disable=unexpected-keyword-arg
        else:
            output = np.concatenate([[mag], obs.m / mag], dtype=np.float32)  # pylint: disable=unexpected-keyword-arg

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = corl_get_ureg().Quantity(output, obs.units)

        return d


class RotateVectorToLocalRef3dGlueValidator(BaseMultiWrapperGlueValidator):
    """
    A configuration validator for RotateVectorToLocalRef3d

    Attributes
    ----------
    mode : str
        the rotation mode, either 'euler' or 'quaternion'
    apply_inv : bool
        whether or not to invert the input vector
    """
    mode: str = 'quaternion'
    apply_inv: bool = True

    @validator('mode')
    def mode_is_recognized(cls, v):
        """Validate mode"""
        if v not in ['quaternion', 'euler']:
            raise ValueError('mode indicates an unrecognized format for the rotation; accepts "euler" or "quaternion"')
        return v


class RotateVectorToLocalRef3d(BaseMultiWrapperGlue):
    """Multiwrapped glue that transforms a 3d vector to a local reference frame

    First wrapped glue should be the orientation observable (angle or quaternion)
    Second wrapped glue will be rotated into the local reference frame defined by the orientation
    """

    class Fields:
        """
        Fields in this glue
        """
        DIRECT_OBSERVATION = "direct_observation"

    @staticmethod
    def get_validator() -> typing.Type[RotateVectorToLocalRef3dGlueValidator]:
        return RotateVectorToLocalRef3dGlueValidator

    def get_unique_name(self) -> str:
        return self.glues()[1].get_unique_name() + "_Local_Ref"

    @cached_property
    def observation_prop(self):
        # unit = self.glues()[1].observation_space[self.glues()[1].Fields.DIRECT_OBSERVATION].unit
        unit = ""
        prop = BoxProp(low=[-10000, -10000, -10000], high=[10000, 10000, 10000], unit=unit)
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: prop})

    @cached_property
    def observation_space(self) -> gymnasium.spaces.Space:
        source_glue = self.glues()[1]
        source_obs_space = source_glue.observation_space[source_glue.Fields.DIRECT_OBSERVATION]
        mag = np.linalg.norm(np.maximum(np.abs(source_obs_space.low), np.abs(source_obs_space.high)))
        d = gymnasium.spaces.dict.Dict()
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(low=-mag, high=mag, shape=(3, ), dtype=np.float32)
        return d

    @cached_property
    def normalized_observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        return self.observation_space

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        orientation_wrapped = self.glues()[0]
        input_vector_wrapped = self.glues()[1]

        orientation_obs = orientation_wrapped.get_observation(other_obs, obs_space,
                                                              obs_units)[orientation_wrapped.Fields.DIRECT_OBSERVATION]
        input_vector = input_vector_wrapped.get_observation(other_obs, obs_space, obs_units)[input_vector_wrapped.Fields.DIRECT_OBSERVATION]

        if self.config.mode == 'euler':
            orientation = Rotation.from_euler('z', angles=orientation_obs.m[0])
        elif self.config.mode == 'quaternion':
            orientation = Rotation.from_quat(orientation_obs.m)

        if self.config.apply_inv:
            rotated_vector = orientation.inv().apply(input_vector.m).astype(np.float32)
        else:
            rotated_vector = orientation.apply(input_vector.m).astype(np.float32)

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = corl_get_ureg().Quantity(rotated_vector, "dimensionless")

        return d


class AngleToUnitVector(BaseWrapperGlue):
    """Wrapper glue that converts an angle about the z axis into a unit vector"""

    class Fields:
        """
        Fields in this glue
        """
        DIRECT_OBSERVATION = "direct_observation"

    def get_unique_name(self) -> str:
        return self.glue().get_unique_name() + "_AngleToUnitVector"

    @cached_property
    def observation_space(self) -> gymnasium.spaces.Space:
        d = gymnasium.spaces.dict.Dict()
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(shape=(3, ), low=-1.0, high=1.0, dtype=np.float32)
        return d

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        obs = self.glue().get_observation(other_obs, obs_space, obs_units)[self.glue().Fields.DIRECT_OBSERVATION]

        sun_position = corl_get_ureg().Quantity(np.array([np.cos(obs[0].m), -np.sin(obs[0].m), 0.]), "m")

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = sun_position

        return d



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains the base Simulator class used by the safe_autonomy_sims team's CWH and Dubins simulators.
"""
# pylint: disable=W0123,W0611,W0221

import abc
import typing
from graphlib import CycleError, TopologicalSorter

from corl.simulators.base_simulator import BaseSimulator, BaseSimulatorResetValidator, BaseSimulatorValidator
from corl.simulators.base_simulator_state import BaseSimulatorState
from pydantic import BaseModel, validator
from pydantic.types import PyObject
from safe_autonomy_simulation.entities import Entity

from safe_autonomy_sims.simulators.initializers.initializer import (
    Accessor,
    EntityAttributeAccessor,
    SimAttributeAccessor,
    StripUnitsInitializer,
)
from safe_autonomy_sims.utils import KeyCollisionError


class SafeRLSimulatorValidator(
    BaseSimulatorValidator,
):
    """
    A validator for the SafeRLSimulator config.
    """


class InitializerResetValidator(BaseModel):
    """
    A configuration validator for the Initializer reset config.

    Attributes
    ----------
    functor: str
        The class module of the Initializer to be instantiated.
    config: dict
        The dict of values to be passed to the Initializer's construtor.
    """
    functor: PyObject
    config: typing.Dict = {}


class AgentResetParamsValidator(BaseModel):
    """A configuration validator for Agent platform reset parameters

    Attributes
    ----------
    initializer: InitializerResetValidator
        params for bulding an initializer object
    config: Dict
        Dictionary of initialization parameters to pass to the entity's initializer/constructor
    """
    initializer: typing.Optional[InitializerResetValidator] = None
    config: typing.Dict = {}


class AdditionalEntityValidator(BaseModel):
    """A configuration validator for additional sim entities

    Attributes
    ----------
    platform: str
        Name of platform type that can be found in sim's platform map
        Uses this value to extract the entity class associated with this platform
        Do no use with entity_class parameter. Use only one.
    entity_class: PyObject or str
        Python class of entity object. Can be specified using python identifier string
        Do not use with platform parameter. Use only one.
    initializer: InitializerResetValidator
        params for bulding an initializer object
    config: Dict
        Dictionary of initialization parameters to pass to the entity's initializer/constructor
    """
    platform: typing.Optional[str] = None
    entity_class: typing.Optional[PyObject] = None
    initializer: typing.Optional[InitializerResetValidator] = None
    config: typing.Dict[str, typing.Any] = {}

    @validator("entity_class")
    def entity_class_platform_xor(cls, entity_class, values):
        """validator to make sure one and exactly one of platform or entity_class is used"""
        platform = values['platform']
        if entity_class is None and platform is None:
            raise ValueError("Must specify platform or entity_class")
        if entity_class is not None and platform is not None:
            raise ValueError("Do not specify both platform and entity_class. Use one parameter only.")
        return entity_class


class SafeRLSimulatorResetValidator(BaseSimulatorResetValidator):
    """
    A configuration validator for the SafeRLSimulator reset method.

    Attributes
    ----------
    platforms: dict
        Contains individual initialization dicts for each agent.
        Key is platform name, value is platform's initialization dict.
    initializer: InitializerResetValidator
        Optionally, the user can define the functor and config for a BaseInitializer class, which modifies
        the agent-specific initialization dicts found in platforms.
    additional_entities: dict
        Contains individual initialization dicts for additional, non-agent simulation entities
        Key is entity name, value is entity's initialization dict.
    """
    platforms: typing.Optional[typing.Dict[str, typing.Dict]] = {}
    # StripUnitsInitializer is the default as package dependencies could be using Pint and CoRL doesn't
    # use Pint
    default_initializer: InitializerResetValidator = InitializerResetValidator(functor=StripUnitsInitializer)
    additional_entities: typing.Dict[str, AdditionalEntityValidator] = {}
    init_state: typing.Optional[BaseModel] = None


class SafeRLSimulatorState(BaseSimulatorState):
    """
    The basemodel for the state of the SafeRLSimulator.

    Attributes
    ----------
    dict
        simulation entities
    """
    sim_entities: typing.Dict


class SafeRLSimulator(BaseSimulator):
    """
    The base simulator class used by the CWH and Dubins simulators. SafeRLSimulator is responsible for
    initializing the platform objects for a simulation and knowing how to set up episodes based on input
    parameters from a parameter provider. It is also responsible for reporting the simulation state at
    each timestep.
    """

    @property
    def get_simulator_validator(self) -> typing.Type[SafeRLSimulatorValidator]:
        return SafeRLSimulatorValidator

    @property
    def get_reset_validator(self) -> typing.Type[SafeRLSimulatorResetValidator]:
        return SafeRLSimulatorResetValidator

    @property
    def step_size(self) -> float:
        """Simulator step size in seconds"""
        return 1 / self.frame_rate

    @property
    def sim_time(self) -> float:
        return self.clock

    @property
    def platforms(self) -> typing.List:
        return list(self._state.sim_platforms)

    def __init__(self, **kwargs: dict[str, typing.Any]):
        self.config: SafeRLSimulatorValidator
        super().__init__(**kwargs)
        self.agent_sim_entities: typing.Dict[str, Entity] = {}
        self.additional_sim_entities: typing.Dict[str, Entity] = {}
        self.platform_map = self._construct_platform_map()
        self.sim_entities: typing.Dict[str, Entity] = {}
        self.clock = 0.0
        self.last_entity_actions: typing.Dict[typing.Any, typing.Any] = {}
        self.sim_platforms = self.construct_platforms()
        self.init_state = None

        self._state: SafeRLSimulatorState

    def reset(self, config):
        reset_config = self.get_reset_validator(**config)
        self.init_state = reset_config.init_state

        self.clock = 0.0
        self.last_entity_actions = {}

        entity_init_map = self._construct_entity_init_map(reset_config)
        self.sim_entities = self._construct_sim_entities(reset_config, entity_init_map)
        self.sim_platforms = self.construct_platforms()

        self._state = self._construct_simulator_state()
        self.update_sensor_measurements()
        return self._state

    def _construct_entity_init_map(self, reset_config: SafeRLSimulatorResetValidator):
        entity_init_map = {}
        for agent_id, agent_config in self.config.agent_configs.items():
            platform_config = agent_config.platform_config
            sim_config_kwargs = platform_config.get("kwargs", {})

            if reset_config.platforms is None:
                agent_reset_config = AgentResetParamsValidator()
            else:
                agent_reset_config = AgentResetParamsValidator(**reset_config.platforms.get(agent_id, {}))

            agent_init_params = {**sim_config_kwargs, **agent_reset_config.config}

            entity_class = self.platform_map[platform_config.get('platform', 'default')][0]

            if agent_reset_config.initializer is None:
                entity_initializer = reset_config.default_initializer.functor(reset_config.default_initializer.config)
            else:
                entity_initializer = agent_reset_config.initializer.functor(agent_reset_config.initializer.config)

            entity_init_map[agent_id] = {
                'class': entity_class,
                'initializer': entity_initializer,
                'params': agent_init_params,
            }

        for entity_name, entity_config in reset_config.additional_entities.items():
            if entity_name in entity_init_map:
                KeyCollisionError(entity_name, f"additional entity name collision: '{entity_name}' is used twice")

            if entity_config.platform is not None:
                entity_class = self.platform_map[entity_config.platform][0]
            else:
                entity_class = entity_config.entity_class

            if entity_config.initializer is None:
                entity_initializer = reset_config.default_initializer.functor(reset_config.default_initializer.config)
            else:
                entity_initializer = entity_config.initializer.functor(entity_config.initializer.config)

            entity_init_map[entity_name] = {
                'class': entity_class,
                'initializer': entity_initializer,
                'params': entity_config.config,
            }

        entity_init_map = self._construct_entity_accessors(entity_init_map)

        return entity_init_map

    def _construct_entity_accessors(self, entity_init_map):
        for _, entity_map_items in entity_init_map.items():
            for param_name, param in entity_map_items['params'].items():
                if isinstance(param, dict) and "accessor" in param:
                    accessor_key = param["accessor"]
                    if accessor_key == "entity_attribute":
                        accessor_obj = EntityAttributeAccessor(**param.get('config', {}))
                    elif accessor_key == "sim_attribute":
                        accessor_obj = SimAttributeAccessor(**param.get('config', {}))
                    else:
                        raise ValueError(f"Invalid accessor {accessor_key}")
                    entity_map_items['params'][param_name] = accessor_obj

        return entity_init_map

    def _generate_entity_init_order(self, entity_init_map) -> typing.List[str]:

        # key is node name, value is set of predecessors (i.e. dependencies)
        dependency_graph: typing.Dict[str, typing.Set] = {entity_name: set() for entity_name in entity_init_map}

        for entity_name, entity_map_items in entity_init_map.items():
            for _, param in entity_map_items['params'].items():
                if isinstance(param, Accessor):
                    dependency_graph[entity_name] = dependency_graph[entity_name].union(param.dependencies)

        try:
            sorter = TopologicalSorter(graph=dependency_graph)
            init_order = list(sorter.static_order())
        except CycleError as e:
            raise ValueError(
                "Cycle detected among entity initializer accessors."
                "Make sure that entities with accessors don't depend on each other"
            ) from e

        return init_order

    def _resolve_accessors(self, params, sim_entities):
        for param_name, param in params.items():
            if isinstance(param, Accessor):
                params[param_name] = param.access(self, sim_entities)

        return params

    @abc.abstractmethod
    def _construct_platform_map(self) -> dict:
        ...

    def _construct_simulator_state(self) -> dict:
        return SafeRLSimulatorState(sim_platforms=self.sim_platforms, sim_time=self.clock, sim_entities=self.sim_entities)

    def _construct_sim_entities(
        self,
        reset_config: SafeRLSimulatorResetValidator,  # pylint: disable=unused-argument
        entity_init_map: typing.Dict[str, typing.Dict]
    ) -> typing.Dict[str, Entity]:
        """Constructs the simulator

        Parameters
        ----------
        reset_config : SafeRLSimulatorResetValidator
            reset config validator
        entity_init_map: dict
            map of per entity initialization items. Each entity name is a key with a 'class', 'initializer', and 'params' nested dict items

        Returns
        -------
        dict[str: sim_entity]
            Dictionary mapping entity id to simulation backend entity.
        """

        init_order = self._generate_entity_init_order(entity_init_map)

        sim_entities: typing.Dict[str, typing.Any] = {}
        for entity_name in init_order:
            entity_init_items = entity_init_map[entity_name]
            params = entity_init_items['params']
            params = self._resolve_accessors(params, sim_entities)
            transformed_params = entity_init_items['initializer'](**params)
            entity = entity_init_items['class'](name=entity_name, **transformed_params)
            sim_entities[entity_name] = entity

        self.agent_sim_entities = {}
        self.additional_sim_entities = {}
        for entity_name, entity in sim_entities.items():
            if entity_name in self.config.agent_configs:
                self.agent_sim_entities[entity_name] = entity
            else:
                self.additional_sim_entities[entity_name] = entity

        return sim_entities

    def construct_platforms(self) -> dict:
        """
        Gets the platform object associated with each simulation entity.

        Returns
        -------
        tuple
            Collection of platforms associated with each simulation entity.
        """
        sim_platforms = {}
        for platform_id, entity in self.agent_sim_entities.items():
            agent_config = self.config.agent_configs[platform_id]
            platform_config = agent_config.platform_config
            platform_class = self.platform_map[platform_config.get('platform', 'default')][1]
            sim_platforms[platform_id] = platform_class(platform_name=platform_id, platform=entity, parts_list=agent_config.parts_list)
        return sim_platforms

    def update_sensor_measurements(self):
        """
        Update and cache all the measurements of all the sensors on each platform.
        """
        for plat in self._state.sim_platforms.values():
            for sensor in plat.sensors.values():
                sensor.calculate_and_cache_measurement(state=self._state)

    def mark_episode_done(self, done_info: typing.OrderedDict, episode_state: typing.OrderedDict, metadata: dict | None = None):
        """
        Takes in the done_info specifying how the episode completed
        and does any book keeping around ending an episode

        Parameters
        ----------
        done_info : OrderedDict
            The Dict describing which Done conditions ended an episode
        episode_state : OrderedDict
            The episode state at the end of the simulation
        """

    def save_episode_information(self, dones, rewards, observations, observation_units):
        pass

    def step(self, platforms_to_action):
        step_size = self.step_size
        self._step_entity_state(step_size=step_size, platforms_to_action=platforms_to_action)
        self._step_update_time(step_size=step_size)
        self._step_update_sim_statuses(step_size=step_size)
        self.update_sensor_measurements()
        return self._state

    def _step_entity_state(self, step_size: float, platforms_to_action: typing.Set[str]):
        entity_actions = self._step_get_entity_actions(step_size=step_size, platforms_to_action=platforms_to_action)

        for entity_name, entity in self.sim_entities.items():
            action = entity_actions.get(entity_name, None)
            entity.add_control(action)
            entity.step(step_size=step_size)

    def _step_get_entity_actions(
        self,
        step_size: float,  # pylint: disable = unused-argument
        platforms_to_action: typing.Set[str]
    ) -> typing.Dict:
        entity_actions = {}
        for platform in self._state.sim_platforms.values():
            platform_id = platform.name
            if platform_id in platforms_to_action:
                action = platform.get_applied_action().m
                entity_actions[platform_id] = action
                self.last_entity_actions[platform_id] = action
            else:
                entity_actions[platform_id] = self.last_entity_actions[platform_id]
        return entity_actions

    def _step_update_time(self, step_size: float):
        self.clock += step_size
        self._state.sim_time = self.clock  # pylint: disable=W0201
        for platform in self._state.sim_platforms.values():
            platform.sim_time = self.clock

    def _step_update_sim_statuses(self, step_size: float):
        """perform custom updates on derived simulation status properties"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains various utility functions.
"""
import typing

import numpy as np
from corl.simulators.base_platform import BasePlatform
from corl.simulators.common_platform_utils import get_sensor_by_name
from pydantic import BaseModel


def shallow_dict_merge(a: typing.Dict, b: typing.Dict, in_place: bool = False, allow_collisions: bool = True):
    """
    Merges dictionaries a and b. Keeps entries from input a if collisions as ignored.

    This is a shallow merge, values in a and b are not copied

    Parameters
    ----------
    a : dict
        Merged into output dictionary. Values preferred in key collision if allow_collisions is false. Modified in-place if in_place True
    b : dict
        Merged into output dictionary. Values ignored if key collides with a and allow_collisions is false.
    in_place: bool
        Modifies input dict a in-place if True, by default False
    allow_collisions : bool, optional
        If True, takes values from input dict a when keys collide.
        If False, raises KeyCollisionError when keys collide.chief_position
        Formed by merging dictionaries a and b
    """

    if in_place:
        output = a
    else:
        output = {**a}

    for k, v in b.items():
        if k in output:
            if allow_collisions:
                continue

            raise KeyCollisionError(k)

        output[k] = v

    return output


class KeyCollisionError(Exception):
    """Custom exception for dictionary merging key collision errors

    Parameters
    ----------
    key : Any
        colliding key
    message: str
        Optional error message
    """

    def __init__(self, key, message: str = None):
        self.key = key
        if message is None:
            message = f"Keys '{key}' collided"
        super().__init__(message)


def get_relative_position(platform: BasePlatform, reference_position_sensor_name: str):
    """
    Finds the relative position between a platform and its reference_position Sensor's returned position.

    Parameters
    ----------
    platform: BasePlatform
        The name of the platform whose relative position will be returned
    reference_position_sensor_name: str
        The name of the sensor on the platform responsible for tracking the absolute position of a reference entity

    Returns
    -------
    relative_position: numpy.ndarray
        The x,y,z length vector describing the relative position of the platform from the reference entity.
    """

    position = platform.position
    reference_position = get_sensor_by_name(platform, reference_position_sensor_name).get_measurement()
    relative_position = np.array(position) - reference_position.m

    return relative_position


def get_relative_velocity(platform: BasePlatform, reference_velocity_sensor_name: str):
    """
    Finds the relative velocity between a platform and its reference_position Sensor's returned velocity.

    Parameters
    ----------
    platform: BasePlatform
        The name of the platform whose relative position will be returned
    reference_velocity_sensor_name: str
        The name of the sensor on the platform responsible for tracking the absolute velocity of a reference entity

    Returns
    -------
    relative_velocity: numpy.ndarray
        The x,y,z length vector describing the relative velocity of the platform from the reference entity.
    """

    velocity = platform.velocity
    reference_velocity = get_sensor_by_name(platform, reference_velocity_sensor_name).get_measurement()
    relative_velocity = np.array(velocity) - reference_velocity.m

    return relative_velocity


def velocity_limit(distance, velocity_threshold, threshold_distance, mean_motion, slope=2.0):
    """
    Get the velocity limit from the platform's current distance from the docking region.

    Parameters
    ----------
    distance: numpy.ndarray
        The x,y,z length vector describing the distance of the platform from the docking region.
    velocity_threshold: float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance: float
        The radius of the docking region.
    mean_motion: float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    slope: float
        The slope of the linear velocity limit as a function of distance from docking region.

    Returns
    -------
    float
        The velocity limit given the platform's position.
    """

    vel_limit = velocity_threshold
    if distance > threshold_distance:
        vel_limit += slope * mean_motion * (distance - threshold_distance)
    return vel_limit


def max_vel_violation(relative_position, relative_velocity, velocity_threshold, threshold_distance, mean_motion, lower_bound, slope=2.0):
    """
    Get the magnitude of a velocity limit violation if one has occurred.

    Parameters
    ----------
    relative_position: numpy.ndarray
        The x,y,z length vector describing the distance of the platform from the docking region.
    relative_velocity: numpy.ndarray
        The x,y,z length vector describing the velocity of the platform with respect to the docking region.
    velocity_threshold: float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance: float
        The radius of the docking region.
    mean_motion: float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    lower_bound: bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    slope: float
        The slope of the linear velocity limit as a function of distance from docking region.

    Returns
    -------
    violated: bool
        Boolean value indicating if the velocity limit has been violated.
    violation: float
        The magnitude of the velocity limit violation.
    """
    distance = np.linalg.norm(relative_position)
    relative_velocity_magnitude = np.linalg.norm(relative_velocity)

    vel_limit = velocity_limit(distance, velocity_threshold, threshold_distance, mean_motion, slope=slope)

    violation = relative_velocity_magnitude - vel_limit
    violated = bool(relative_velocity_magnitude > vel_limit)
    if lower_bound:
        violation *= -1
        violated = bool(relative_velocity_magnitude < vel_limit)

    return violated, violation


class VelocityConstraintValidator(BaseModel):
    """
    A configuration validator for velocity constraint configuration options.

    Attributes
    ----------
    velocity_threshold : float
        The maximum tolerated velocity within crashing region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the crashing region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    """
    velocity_threshold: float
    threshold_distance: float
    mean_motion: float = 0.001027
    lower_bound: bool = False
    slope: float = 2.0


def get_closest_fft_distance(state: np.ndarray, n: float, time_array: typing.Union[np.ndarray, list]) -> float:
    """
    Get the closest Free Flight Trajectory (FFT) distance over a given time horizon.
    Use closed form CWH dynamics to calculate future states in the absence of control

    Parameters
    ----------
    state: np.ndarray
        Initial state
    n: float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    time_array: Union[np.ndarray, list]
        Array containing each point in time to check the FFT trajectory

    Returns
    -------
    float
        Closest relative distance to the origin achieved during the FFT
    """

    distances = []
    for t in time_array:
        x = (4 - 3 * np.cos(n * t)) * state[0] + np.sin(n * t) * state[3] / n + 2 / n * (1 - np.cos(n * t)) * state[4]
        y = 6 * (np.sin(n * t) - n * t) * state[0] + state[1] - 2 / n * (1 - np.cos(n * t)) * state[3] + (4 * np.sin(n * t) -
                                                                                                          3 * n * t) * state[4] / n
        z = state[2] * np.cos(n * t) + state[5] / n * np.sin(n * t)
        distances.append(np.linalg.norm([x, y, z]))
    return float(min(distances))



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import math
import typing

import matplotlib.pyplot as plt
import numpy as np
import yaml
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer

from safe_autonomy_sims.evaluation.animation.base_animation import BaseAnimationModule
from safe_autonomy_sims.evaluation.launch.serialize_cwh3d import SerializeCWH3D
from safe_autonomy_sims.simulators.inspection_simulator import InspectionPoints


class InspectionAnimation(BaseAnimationModule):
    """
    Module for making animations/plots from a checkpoint for the inspection environment

    Parameters
    ----------
    expr_config_path: str
        The absolute path to the experiment config used in training
    task_config_path: str
        The absolute path to the task_config used in training
    platform_serializer_class: PlatformSerializer
        The PlatformSerializer subclass capable of saving the state of the Platforms used in the evaluation environment
    parameters_path: str
        Path to a yml file with parameters
    radius: float
        Radius of the sphere of inspection points
    fft_time: float
        Total time for the Free Flight Trajectory (FFT) calculations in seconds, by default 2 days
    points_algorithm: str
        Points algorithm to use, either "cmu" or "fibonacci"
    """

    def __init__(
        self,
        expr_config_path: str = 'safe-autonomy-sims/configs/translational-inspection/experiment.yml',
        task_config_path: str = 'safe-autonomy-sims/configs/translational-inspection/task.yml',
        platform_serializer_class: PlatformSerializer = SerializeCWH3D,
        parameters_path: str = 'safe-autonomy-sims/configs/translational-inspection/parameters.yml',
        radius: float = 10.,
        fft_time: float = 172800.,
        points_algorithm: str = "cmu",
        **kwargs,
    ):

        with open(parameters_path, 'r', encoding='utf-8') as file:
            parameters = yaml.safe_load(file)

        if 'step_size' in parameters:
            self.step_size = parameters['step_size']
        else:
            self.step_size = 10.

        if 'num_points' in parameters:
            self.num_points = parameters['num_points']
        else:
            self.num_points = 100

        if 'mass' in parameters:
            self.mass = parameters['mass']
        else:
            self.mass = 12

        if 'mean_motion' in parameters:
            self.n = parameters['mean_motion']
        else:
            self.n = 0.001027

        self.radius = radius
        self.fft_time = fft_time
        self.points_algorithm = points_algorithm

        self.lw = 3  # linewidth
        self.bool_array_locations = False
        self.delta_v = 0.0
        self.ffts = np.array([])
        self.temp_elements: list[typing.Any] = []

        # Calculate values
        if self.points_algorithm == "cmu":
            self.points = InspectionPoints.points_on_sphere_cmu(self.num_points, radius)
        elif self.points_algorithm == "fibonacci":
            self.points = InspectionPoints.points_on_sphere_fibonacci(self.num_points, radius)
        else:
            raise ValueError('Points algorithm must be either "cmu" or "fibonacci"')

        super().__init__(
            expr_config_path=expr_config_path,
            task_config_path=task_config_path,
            platform_serializer_class=platform_serializer_class,
            start_index=1,
            **kwargs
        )

    def get_closest_fft(self, pos, vel):
        """Get closest relative position to the origin along a Free Flight Trajectory (FFT)

        Parameters
        ----------
        pos: list
            position vector
        vel: list
            velocity vector
        """
        time = np.arange(0, self.fft_time, self.step_size)
        positions = []
        for t in time:
            x = (4 - 3 * np.cos(self.n * t)) * pos[0] + np.sin(self.n * t
                                                               ) * vel[0] / self.n + 2 / self.n * (1 - np.cos(self.n * t)) * vel[1]
            y = 6 * (np.sin(self.n * t) - self.n * t) * pos[0] + pos[
                1] - 2 / self.n * (1 - np.cos(self.n * t)) * vel[0] + (4 * np.sin(self.n * t) - 3 * self.n * t) * vel[1] / self.n
            z = pos[2] * np.cos(self.n * t) + vel[2] / self.n * np.sin(self.n * t)
            positions.append(np.linalg.norm([x, y, z]))
        return min(positions)

    def get_data_from_dataframe(self, dataframe):
        positions = []
        velocities = []
        sun_vectors = []
        bool_arrays = []
        scores = []
        for state in dataframe["ObservationVector"][0]:
            positions.append(state.data['Obs_Sensor_Position'].value['direct_observation'].value)
            velocities.append(state.data['Obs_Sensor_Velocity'].value['direct_observation'].value)
            try:
                score = state.data['Obs_Sensor_InspectedPointsScore'].value['direct_observation']
            except KeyError:
                score = np.NAN
            scores.append(score)
            try:
                th = state.data['Obs_Sensor_SunAngle'].value['direct_observation'].value[0]
            except KeyError:
                th = 0.
            sun = np.array([np.cos(th), -np.sin(th), 0])
            sun_vectors.append(sun)
            try:
                new_bool = state.data['Obs_Sensor_BoolArray'].value['direct_observation']
                self.bool_array_locations = True
            except KeyError:
                ins_pts = int(state.data['Obs_Sensor_InspectedPoints'].value['direct_observation'].value)
                new_bool = np.zeros(self.num_points)
                new_bool[0:ins_pts] = 1
            bool_arrays.append(new_bool)

        times = np.arange(len(positions)) * self.step_size

        control_vectors = []
        for control in dataframe["ControlVector"][0]:
            Fx = control.data[list(control.data.keys())[0]].value
            Fy = control.data[list(control.data.keys())[1]].value
            Fz = control.data[list(control.data.keys())[2]].value
            control_vectors.append([Fx, Fy, Fz])

        data = {}

        data['times'] = times
        data['positions'] = np.array(positions)
        data['bool_arrays'] = np.array(bool_arrays)
        data['sun_vectors'] = np.array(sun_vectors)
        data['control_vectors'] = np.array(control_vectors)
        data['velocities'] = np.array(velocities)
        data['scores'] = np.array(scores)

        if self.points_algorithm == "cmu":
            data['bool_arrays'] = np.delete(data['bool_arrays'], -1, 1)  # TODO: fix bool array number of points

        return data

    def setup_ric(self, data, max_pos, return_plots):
        """Set up RIC plots.

        Parameters
        ----------
        data : dict
            data to plot
        max_pos : float
            max position value
        return_plots : dict
            plots to modify

        Returns
        -------
        dict
            modified plots in RIC frame
        """
        ax_ri = plt.subplot2grid((5, 4), (0, 0), colspan=2, rowspan=3)
        ax_ri.grid(True)
        ax_ri.set_xlabel('In-Track [m]')
        ax_ri.set_ylabel('Radial [m]')
        ax_ri.set_xlim([-max_pos, max_pos])
        ax_ri.set_ylim([-max_pos, max_pos])
        ax_ri.set_aspect("equal")
        ax_ri.invert_xaxis()
        if self.filetype == 'png':
            ax_ri.plot(max_pos * 2, max_pos * 2, color=plt.cm.cool(0), linewidth=self.lw, label="Deputy Position")
        else:
            ax_ri.plot(max_pos * 2, max_pos * 2, '*', color=plt.cm.cool(0), linewidth=self.lw, label="Deputy Position")
            ax_ri.plot(max_pos * 2, max_pos * 2, linewidth=5, color='y', label="Sun Vector")
        ax_ri.plot(data['positions'][0][1], data['positions'][0][0], 'b*', markersize=10, label="Initial Position")
        if self.bool_array_locations:
            ax_ri.plot(max_pos * 2, max_pos * 2, 'g*', label="Inspected Point")
            ax_ri.plot(max_pos * 2, max_pos * 2, 'r*', label="Uninspected Point")
        else:
            ax_ri.plot(max_pos * 2, max_pos * 2, 'r*', label="Inspection Point")
        ax_ri.legend()
        return_plots['ax_ri'] = ax_ri

        ax_rc = plt.subplot2grid((5, 4), (3, 0), rowspan=2)
        ax_rc.grid(True)
        ax_rc.set_xlabel('Cross-Track [m]')
        ax_rc.set_ylabel('Radial [m]')
        ax_rc.set_xlim([-max_pos, max_pos])
        ax_rc.set_ylim([-max_pos, max_pos])
        ax_rc.set_aspect("equal")
        ax_rc.invert_xaxis()
        ax_rc.plot(data['positions'][0][2], data['positions'][0][0], 'b*', markersize=10)
        return_plots['ax_rc'] = ax_rc

        ax_ic = plt.subplot2grid((5, 4), (3, 1), rowspan=2)
        ax_ic.grid(True)
        ax_ic.set_xlabel('In-Track [m]')
        ax_ic.set_ylabel('Cross-Track [m]')
        ax_ic.set_xlim([-max_pos, max_pos])
        ax_ic.set_ylim([-max_pos, max_pos])
        ax_ic.set_aspect("equal")
        ax_ic.invert_xaxis()
        ax_ic.invert_yaxis()
        ax_ic.plot(data['positions'][0][1], data['positions'][0][2], 'b*', markersize=10)
        return_plots['ax_ic'] = ax_ic

        return return_plots

    def setup_rp_rv(self, data, return_plots):
        """Set up relative position and velocity plots

        Parameters
        ----------
        data : dict
            data to plot
        return_plots : dict
            plots to modify

        Returns
        -------
        dict
            modified plots with relative position and velocity
        """
        max_rp = []
        for x in data['positions']:
            max_rp.append(np.linalg.norm(x))
        max_rp = max(max_rp)
        self.ffts = []
        for pos, vel in zip(data['positions'], data['velocities']):
            self.ffts.append(self.get_closest_fft(pos, vel))
        max_rp = max(max_rp, max(self.ffts)) * 1.1
        ax_rp = plt.subplot2grid((5, 4), (1, 2), colspan=3)
        ax_rp.set_xlim([0, data['times'][-1]])
        ax_rp.set_ylim([0, max_rp])
        ax_rp.set_xlabel('Time [sec]')
        ax_rp.set_ylabel('Relative Distance [m]')
        ax_rp.grid(True)
        ax_rp.plot(-10, -10, 'r', label=r'Closest FFT', linewidth=self.lw)
        ax_rp.plot(-10, -10, 'b', label=r'Trajectory', linewidth=self.lw)
        ax_rp.legend()
        return_plots['ax_rp'] = ax_rp

        max_rv = []
        for x in data['velocities']:
            max_rv.append(np.linalg.norm(x))
        max_rv = max(max_rv) * 1.1
        ax_rv = plt.subplot2grid((5, 4), (2, 2), colspan=3)
        ax_rv.set_xlim([0, data['times'][-1]])
        ax_rv.set_ylim([0, max_rv])
        ax_rv.set_xlabel('Time [sec]')
        ax_rv.set_ylabel('Relative Speed [m/s]')
        ax_rv.grid(True)
        return_plots['ax_rv'] = ax_rv

        return return_plots

    def setup_plots(self, data):
        return_plots = {}

        max_positions = []
        for x in data['positions']:
            max_positions.append(max(abs(x)))
        max_pos = float(max(max_positions) * 1.1)

        self.delta_v = np.sum(np.abs(data['control_vectors'][0])) / self.mass * self.step_size

        # Set plot values
        plt.figure(figsize=(20, 15.04))
        plt.rcParams.update({'font.size': 20, 'text.usetex': True})

        # Setup plots
        return_plots = self.setup_ric(data, max_pos, return_plots)
        return_plots = self.setup_rp_rv(data, return_plots)

        max_dv = 0
        for x in data['control_vectors']:
            max_dv += np.sum(np.abs(x)) / self.mass * self.step_size
        max_dv = max_dv * 1.1
        ax_dv = plt.subplot2grid((5, 4), (4, 2), colspan=3)
        ax_dv.set_xlim([0, data['times'][-1]])
        ax_dv.set_ylim([0, max_dv])
        ax_dv.set_xlabel('Time [sec]')
        ax_dv.set_ylabel(r'Cumulative $\Delta$V [m/s]')
        ax_dv.grid(True)
        return_plots['ax_dv'] = ax_dv

        ax_ip = plt.subplot2grid((5, 4), (0, 2), colspan=3)
        ax_ip.set_xlim([0, data['times'][-1]])
        ax_ip.set_ylim([0, 100])
        ax_ip.set_xlabel('Time [sec]')
        ax_ip.set_ylabel('Points Inspected')
        ax_ip.grid(True)
        ax_ip.plot(-10, -10, 'b', linewidth=self.lw, label='Percentage')
        if not any(np.isnan(data['scores'])):
            ax_ip.plot(-10, -10, 'r', linewidth=self.lw, label='Score')
        ax_ip.legend()
        return_plots['ax_ip'] = ax_ip

        ax_u = plt.subplot2grid((5, 4), (3, 2), colspan=3)
        ax_u.set_xlim([0, data['times'][-1]])
        y_lim = 1 / self.mass * self.step_size * 1.1
        ax_u.set_ylim([-y_lim, y_lim])
        ax_u.set_xlabel('Time [sec]')
        ax_u.set_ylabel('Thrust [m/s]')
        ax_u.grid(True)
        ax_u.plot(-10, -10, 'b', label=r'$R$', linewidth=self.lw)
        ax_u.plot(-10, -10, 'r', label=r'$I$', linewidth=self.lw)
        ax_u.plot(-10, -10, 'g', label=r'$C$', linewidth=self.lw)
        ax_u.legend()
        return_plots['ax_u'] = ax_u

        return return_plots

    def make_plots_in_loop(self, axes, data, i):
        max_rv = []
        for x in data['velocities']:
            max_rv.append(max(abs(x)))
        max_rv = max(max_rv) * 1.1

        max_positions = []
        for x in data['positions']:
            max_positions.append(max(abs(x)))
        max_pos = float(max(max_positions) * 1.1)

        vel_color = np.linalg.norm(data['velocities'][i]) / (max_rv / 1.1)

        axes['ax_ri'].plot(
            [data['positions'][i - 1][1], data['positions'][i][1]], [data['positions'][i - 1][0], data['positions'][i][0]],
            color=plt.cm.cool(vel_color),
            linewidth=self.lw
        )
        axes['ax_rc'].plot(
            [data['positions'][i - 1][2], data['positions'][i][2]], [data['positions'][i - 1][0], data['positions'][i][0]],
            color=plt.cm.cool(vel_color),
            linewidth=self.lw
        )
        axes['ax_ic'].plot(
            [data['positions'][i - 1][1], data['positions'][i][1]], [data['positions'][i - 1][2], data['positions'][i][2]],
            color=plt.cm.cool(vel_color),
            linewidth=self.lw
        )

        if self.filetype != 'png':
            self.animate_sun_and_points(axes, data, i, max_pos)

        if self.filetype != 'png' or i == len(data['times']) - 1:
            c = 'r'
            # Points on sphere plot
            plot_points = []
            for j, p in enumerate(data['bool_arrays'][i]):
                if self.bool_array_locations:
                    if p:
                        c = 'g'
                    else:
                        c = 'r'

                plot_points.extend(axes['ax_ri'].plot(self.points[j][1], self.points[j][0], '*', color=c))
                plot_points.extend(axes['ax_rc'].plot(self.points[j][2], self.points[j][0], '*', color=c))
                plot_points.extend(axes['ax_ic'].plot(self.points[j][1], self.points[j][2], '*', color=c))
            self.temp_elements.append(plot_points)

        last_delta_v = self.delta_v
        self.delta_v += np.sum(np.abs(data['control_vectors'][i])) / self.mass * self.step_size
        axes['ax_dv'].plot([data['times'][i - 1], data['times'][i]], [last_delta_v, self.delta_v], 'b', linewidth=self.lw)

        last_rel_pos = np.linalg.norm(data['positions'][i - 1])
        rel_pos = np.linalg.norm(data['positions'][i])
        axes['ax_rp'].plot([data['times'][i - 1], data['times'][i]], [self.ffts[i - 1], self.ffts[i]], 'r', linewidth=self.lw)
        axes['ax_rp'].plot([data['times'][i - 1], data['times'][i]], [last_rel_pos, rel_pos], 'b', linewidth=self.lw)

        last_rel_vel = np.linalg.norm(data['velocities'][i - 1])
        rel_vel = np.linalg.norm(data['velocities'][i])
        axes['ax_rv'].plot(
            [data['times'][i - 1], data['times'][i]], [last_rel_vel, rel_vel], color=plt.cm.cool(vel_color), linewidth=self.lw
        )

        last_control_dv = data['control_vectors'][i - 1] / self.mass * self.step_size
        control_dv = data['control_vectors'][i] / self.mass * self.step_size
        axes['ax_u'].plot([data['times'][i - 1], data['times'][i]], [last_control_dv[0], control_dv[0]], 'b', linewidth=self.lw)
        axes['ax_u'].plot([data['times'][i - 1], data['times'][i]], [last_control_dv[1], control_dv[1]], 'r', linewidth=self.lw)
        axes['ax_u'].plot([data['times'][i - 1], data['times'][i]], [last_control_dv[2], control_dv[2]], 'g', linewidth=self.lw)

        last_pts = sum(data['bool_arrays'][i - 1]) / self.num_points * 100
        pts = sum(data['bool_arrays'][i]) / self.num_points * 100
        axes['ax_ip'].plot([data['times'][i - 1], data['times'][i]], [last_pts, pts], 'b', linewidth=self.lw)
        if not any(np.isnan(data['scores'])):
            axes['ax_ip'].plot(
                [data['times'][i - 1], data['times'][i]], [data['scores'][i - 1] * 100, data['scores'][i] * 100], 'r', linewidth=self.lw
            )

    def animate_sun_and_points(self, axes, data, i, max_pos):
        """Animate the sun vector and points

        Parameters
        ----------
        axes : pyplot.Axes
            Plot axes
        data : dict
            data to plot
        i : int
            data index
        max_pos : float
            maximum position value
        """
        # Sun vector plot
        sun_vector = data['sun_vectors'][i] / np.linalg.norm(data['sun_vectors'][i])
        point_inSunDir = max_pos / 1.1 * sun_vector
        sun_line1, = axes['ax_ri'].plot([point_inSunDir[1], 0], [point_inSunDir[0], 0], linewidth=5, color='y')
        sun_line2, = axes['ax_rc'].plot([point_inSunDir[2], 0], [point_inSunDir[0], 0], linewidth=5, color='y')
        sun_line3, = axes['ax_ic'].plot([point_inSunDir[1], 0], [point_inSunDir[2], 0], linewidth=5, color='y')
        self.temp_elements.append(sun_line1)
        self.temp_elements.append(sun_line2)
        self.temp_elements.append(sun_line3)

        # Time text boxes
        angle = np.arctan2(sun_vector[1], sun_vector[0])
        satellite_time = (-angle / (2 * np.pi) + 0.5) * 24
        mins, hrs = math.modf(satellite_time)
        mins *= 60
        txt = axes['ax_ri'].text(0.75, -0.1, f"LST: {hrs:02.0f}:{mins:02.0f}", transform=axes['ax_ri'].transAxes)
        mins, hrs = math.modf(data['times'][i] / 3600)
        mins *= 60
        txt2 = axes['ax_ri'].text(0.0, -0.1, f"Time Elapsed: {hrs:02.0f}:{mins:02.0f}", transform=axes['ax_ri'].transAxes)

        self.temp_elements.append(txt)
        self.temp_elements.append(txt2)

    def remove_temp_elements(self, axes):
        for tmp in self.temp_elements:
            if isinstance(tmp, list):
                for p in tmp:
                    p.remove()
            else:
                tmp.remove()

        self.temp_elements = []

    def print_metrics(self, data):
        print('Metrics:')
        print(f"Inspected Points: {np.count_nonzero(data['bool_arrays'][-1]):0.0f}/{len(data['bool_arrays'][-1]):0.0f}")
        print(f"Time: {data['times'][-1]:0.1f} sec")
        print(f"Delta V: {self.delta_v:0.2f} m/s")



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

The `safe-autonomy-sims` package provides reinforement learning environments
built using the CoRL framework.

The included environments define safety-critical spacecraft docking and
spacecraft inspection tasks. The intent of these environments is to
provide a baseline for training RL agents to safely perform these tasks.
"""



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a patchable CoRL experiment for use with multiagent environments.
"""
import argparse
import importlib
import logging
import os
import pathlib
import re
import socket
import sys
import typing
import warnings
from datetime import datetime

import git
import ray
from corl.environment.default_env_rllib_callbacks import EnvironmentDefaultCallbacks
from corl.environment.multi_agent_env import ACT3MultiAgentEnv, ACT3MultiAgentEnvValidator
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.episode_parameter_providers.remote import RemoteEpisodeParameterProvider
from corl.experiments.base_experiment import BaseExperiment, BaseExperimentValidator
from corl.experiments.rllib_utils.policy_mapping_functions import PolicyIsAgent
from corl.libraries.factory import Factory
from corl.parsers.yaml_loader import apply_patches
from corl.policies.base_policy import BasePolicyValidator
from pydantic import BaseModel, validator
from pydantic.types import PyObject
from ray import tune
from ray.rllib.algorithms.callbacks import make_multi_callbacks
from ray.rllib.env.env_context import EnvContext
from ray.tune.registry import get_trainable_cls


class RllibPolicyMappingValidator(BaseModel):
    """A configuration validator used to generate policy mappings

    Attributes
    ----------
    functor: PyObject
        policy mapping functor
    config: dict
        configuration parameters for policy mapping functor
    """
    functor: PyObject = PolicyIsAgent
    config: dict = {}


class SafeAutonomyRllibExperimentValidator(BaseExperimentValidator):
    """
    A configuration validator for SafeAutonomyRllibExperiment.

    Attributes
    ----------
    ray_config : dict
        dictionary to be fed into ray init, validated by ray init call
    env_config : dict
        environment configuration, validated by environment class
    rllib_configs : dict
        a mapping of compute platforms to rllib configs, see apply_patches_rllib_configs
        for information on the typing
    tune_config : dict
        kwarg arguments to be sent to tune for this experiment
    trainable_config : dict
        training configuration parameters
    hparam_search_class : PyObject
        rllib hyperparameter search object
    hparam_search_config : PyObject
        configuration parameters for hyperparameter search
    extra_callbacks : list
        extra rllib callbacks that will be added to the callback list
    trial_creator_function : PyObject
        this function will overwrite the default trial string creator
        and allow more fine tune trial name creators
    policy_mapping : RllibPolicyMappingValidator
        policy mapping function and configuration options
    """
    ray_config: typing.Dict[str, typing.Any]
    env_config: EnvContext
    rllib_configs: typing.Dict[str, typing.Dict[str, typing.Any]]
    tune_config: typing.Dict[str, typing.Any]
    trainable_config: typing.Optional[typing.Dict[str, typing.Any]]
    hparam_search_class: typing.Optional[PyObject]
    hparam_search_config: typing.Optional[typing.Dict[str, typing.Any]]
    extra_callbacks: typing.Optional[typing.List[PyObject]]
    trial_creator_function: typing.Optional[PyObject]
    policy_mapping: RllibPolicyMappingValidator = RllibPolicyMappingValidator()

    @validator('rllib_configs', pre=True)
    def apply_patches_rllib_configs(cls, v):  # pylint: disable=no-self-argument
        """
        The dictionary of rllib configs may come in as a dictionary of
        lists of dictionaries, this function is responsible for collapsing
        the list down to a typing.Dict[str, typing.Dict[str, typing.Any]]
        instead of
        typing.Dict[str, typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]
        """
        if not isinstance(v, dict):
            raise RuntimeError("rllib_configs are expected to be a dict of keys to different compute configs")
        rllib_configs = {}
        for key, value in v.items():
            if isinstance(value, list):
                rllib_configs[key] = apply_patches(value)
            elif isinstance(value, dict):
                rllib_configs[key] = value
        return rllib_configs

    @validator('ray_config', 'tune_config', 'trainable_config', 'env_config', pre=True)
    def apply_patches_configs(cls, v):  # pylint: disable=no-self-argument
        """
        reduces a field from
        typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]
        to typing.Dict[str, typing.Any]
        by patching the first dictionary in the list with each patch afterwards
        """
        if isinstance(v, list):
            v = apply_patches(v)
        return v

    @validator('env_config')
    def no_horizon(cls, v):
        """Ensure that the horizon is not specified in the env_config."""
        if 'horizon' in v:
            raise ValueError('Cannot specify the horizon in the env_config')
        return v


class RllibPolicyValidator(BasePolicyValidator):
    """
    A configuration validator for RllibPolicy

    Attributes
    ----------
    policy_class: PyObject
        callable policy class, if None will use default from trainer
    train: bool
        should this policy be trained
    """
    config: typing.Dict[str, typing.Any] = {}
    policy_class: typing.Union[PyObject, None] = None
    train: bool = True


class SafeAutonomyRllibExperiment(BaseExperiment):
    """
    This RllibExperiment is an experiment for running
    multiagent configurable environments with patchable settings
    """

    def __init__(self, **kwargs) -> None:
        self.config: SafeAutonomyRllibExperimentValidator
        self._logger = logging.getLogger(SafeAutonomyRllibExperiment.__name__)
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> typing.Type[SafeAutonomyRllibExperimentValidator]:
        return SafeAutonomyRllibExperimentValidator

    @property
    def get_policy_validator(self) -> typing.Type[RllibPolicyValidator]:
        """Return validator"""
        return RllibPolicyValidator

    def run_experiment(self, args: argparse.Namespace) -> None:

        rllib_config = self._select_rllib_config(args.compute_platform)

        if args.compute_platform in ['ray']:
            self._update_ray_config_for_ray_platform()

        self._add_trial_creator()

        # This needs to be before the ray cluster is initialized
        if args.debug:
            self.config.ray_config['local_mode'] = True

        ray.init(**self.config.ray_config)

        self.config.env_config["agents"], self.config.env_config["agent_platforms"] = self.create_agents(
            args.platform_config, args.agent_config
        )

        self.config.env_config["horizon"] = rllib_config["horizon"]

        if args.output:
            self.config.env_config["output_path"] = args.output
            self.config.tune_config["local_dir"] = args.output

        if args.name:
            self.config.env_config["TrialName"] = args.name

        if args.other_platform:
            self.config.env_config["other_platforms"] = self.create_other_platforms(args.other_platform)

        if not self.config.ray_config['local_mode']:
            self.config.env_config['episode_parameter_provider'] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                Factory(**self.config.env_config['episode_parameter_provider']),
                actor_name=ACT3MultiAgentEnv.episode_parameter_provider_name
            )

            for agent_name, agent_configs in self.config.env_config['agents'].items():
                agent_configs.class_config.config['episode_parameter_provider'] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                    Factory(**agent_configs.class_config.config['episode_parameter_provider']), agent_name
                )

        self.config.env_config['epp_registry'] = ACT3MultiAgentEnvValidator(**self.config.env_config).epp_registry

        tmp = ACT3MultiAgentEnv(self.config.env_config)
        tmp_as = tmp.action_space
        tmp_os = tmp.observation_space
        tmp_ac = self.config.env_config['agents']

        policies = {
            policy_name: (
                tmp_ac[policy_name].policy_config["policy_class"],
                policy_obs,
                tmp_as[policy_name],
                tmp_ac[policy_name].policy_config["config"]
            )
            for policy_name,
            policy_obs in tmp_os.spaces.items()
            if tmp_ac[policy_name]
        }

        train_policies = [policy_name for policy_name in policies.keys() if tmp_ac[policy_name].policy_config["train"]]

        self._update_rllib_config(rllib_config, train_policies, policies, args)

        self._enable_episode_parameter_provider_checkpointing()

        if args.profile:
            if "stop" not in self.config.tune_config:
                self.config.tune_config["stop"] = {}
            self.config.tune_config["stop"]["training_iteration"] = args.profile_iterations

        search_class = None
        if self.config.hparam_search_class is not None:
            if self.config.hparam_search_config is not None:
                search_class = self.config.hparam_search_class(**self.config.hparam_search_config)
            else:
                search_class = self.config.hparam_search_class()
            search_class.add_algorithm_hparams(rllib_config, self.config.tune_config)

        self.config = self._add_experiment_output_dir(self.config)

        tune.run(
            config=rllib_config,
            **self.config.tune_config,
        )

    def _update_rllib_config(self, rllib_config, train_policies, policies, args: argparse.Namespace) -> None:
        """
        Update several rllib config fields
        """

        rllib_config["multiagent"] = {
            "policies": policies,
            "policy_mapping_fn": self.config.policy_mapping.functor(**self.config.policy_mapping.config),
            "policies_to_train": train_policies
        }

        rllib_config["env"] = ACT3MultiAgentEnv
        callback_list = [self.get_callbacks()]
        if self.config.extra_callbacks:
            callback_list.extend(self.config.extra_callbacks)  # type: ignore[arg-type]
        rllib_config["callbacks"] = make_multi_callbacks(callback_list)
        rllib_config["env_config"] = self.config.env_config
        now = datetime.now()
        rllib_config["env_config"]["output_date_string"] = f"{now.strftime('%Y%m%d_%H%M%S')}_{socket.gethostname()}"
        rllib_config["create_env_on_driver"] = True
        rllib_config["batch_mode"] = "complete_episodes"

        self._add_git_hashes_to_config(rllib_config)

        if args.debug:
            rllib_config['num_workers'] = 0

    def _add_git_hashes_to_config(self, rllib_config) -> None:
        """adds git hashes (or package version information if git information
        is unavailable) of key modules to rllib_config["env_config"]["git_hash"].
        Key modules are the following:
          - corl,
          - whatever cwd is set to at the time of the function call
            (notionally /opt/project / -rl-agents)
          - any other modules listed in rllib_config["env_config"]["plugin_paths"]

        This information is not actually used by ACT3MultiAgentEnv;
        however, putting it in the env_config means that this
        information is saved to the params.pkl and thus is available
        for later inspection while seeking to understand the
        performance of a trained model.
        """
        try:
            # pattern used below to find root repository paths
            repo_pattern = r"(?P<repopath>.*)\/__init__.py"
            rp = re.compile(repo_pattern)

            corl_pattern = r"corl.*"
            cp0 = re.compile(corl_pattern)
            rllib_config["env_config"]["git_hash"] = {}

            # store hash on cwd
            cwd = os.getcwd()
            try:
                githash = git.Repo(cwd, search_parent_directories=True).head.object.hexsha
                rllib_config["env_config"]["git_hash"]["cwd"] = githash
                self._logger.info(f"cwd hash: {githash}")
            except git.InvalidGitRepositoryError:
                self._logger.warning("cwd is not a git repo\n")

            # go ahead and strip out corl related things from plugin_path
            plugpath = []
            for item in rllib_config['env_config']['plugin_paths']:
                match0 = cp0.match(item)
                if match0 is None:
                    plugpath.append(item)

            plugpath.append('corl')

            # add git hashes to env_config dictionary
            for module0 in plugpath:
                env_hash_key = module0
                module1 = importlib.import_module(module0)
                modulefile = module1.__file__
                if modulefile is not None:
                    match0 = rp.match(modulefile)
                    if match0 is not None:
                        repo_path = match0.group('repopath')
                        try:
                            githash = git.Repo(repo_path, search_parent_directories=True).head.object.hexsha
                            rllib_config["env_config"]["git_hash"][env_hash_key] = githash
                            self._logger.info(f"{module0} hash: {githash}")
                        except git.InvalidGitRepositoryError:
                            # possibly installed in image but not a git repo
                            # look for version number
                            if hasattr(module1, 'version') and hasattr(module1.version, '__version__'):
                                githash = module1.version.__version__
                                rllib_config["env_config"]["git_hash"][env_hash_key] = githash
                                self._logger.info(f"{module0} hash: {githash}")
                            else:
                                self._logger.warning((f"module: {module0}, repopath: {repo_path}"
                                                      "is invalid git repo\n"))
                                sys.stderr.write((f"module: {module0}, repopath: {repo_path}"
                                                  "is invalid git repo\n"))
        except ValueError:
            warnings.warn("Unable to add the gitlab hash to experiment!!!")

    def get_callbacks(self) -> typing.Type[EnvironmentDefaultCallbacks]:
        """Get the environment callbacks"""
        return EnvironmentDefaultCallbacks

    def _select_rllib_config(self, platform: typing.Optional[str]) -> typing.Dict[str, typing.Any]:
        """Extract the rllib config for the proper computational platform

        Parameters
        ----------
        platform : typing.Optional[str]
            Specification of the computational platform to use, such as "local", "hpc", etc.  This must be present in the rllib_configs.
            If None, the rllib_configs must only have a single entry.

        Returns
        -------
        typing.Dict[str, typing.Any]
            Rllib configuration for the desired computational platform.

        Raises
        ------
        RuntimeError
            The requested computational platform does not exist or None was used when multiple platforms were defined.
        """
        if platform is not None:
            return self.config.rllib_configs[platform]

        if len(self.config.rllib_configs) == 1:
            return self.config.rllib_configs[next(iter(self.config.rllib_configs))]

        raise RuntimeError(f'Invalid rllib_config for platform "{platform}"')

    def _update_ray_config_for_ray_platform(self) -> None:
        """Update the ray configuration for ray platforms
        """
        self.config.ray_config['address'] = 'auto'
        self.config.ray_config['log_to_driver'] = False

    def _enable_episode_parameter_provider_checkpointing(self) -> None:

        base_trainer = self.config.tune_config["run_or_experiment"]

        trainer_class = get_trainable_cls(base_trainer)

        class EpisodeParameterProviderSavingTrainer(trainer_class):  # type: ignore[valid-type, misc]
            """
            Tune Trainer that adds capability to restore
            progress of the EpisodeParameterProvider on restoring training
            progress
            """

            def save_checkpoint(self, checkpoint_path):
                """
                adds additional checkpoint saving functionality
                by also saving any episode parameter providers
                currently running
                """
                tmp = super().save_checkpoint(checkpoint_path)

                checkpoint_folder = pathlib.Path(checkpoint_path)

                # Environment
                epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
                env = self.workers.local_worker().env
                epp: EpisodeParameterProvider = env.config.epp
                epp.save_checkpoint(checkpoint_folder / epp_name)

                # Agents
                for agent_name, agent_configs in env.agent_dict.items():
                    epp = agent_configs.config.epp
                    epp.save_checkpoint(checkpoint_folder / agent_name)

                return tmp

            def load_checkpoint(self, checkpoint_path):
                """
                adds additional checkpoint loading functionality
                by also loading any episode parameter providers
                with the checkpoint
                """
                super().load_checkpoint(checkpoint_path)

                checkpoint_folder = pathlib.Path(checkpoint_path).parent

                # Environment
                epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
                env = self.workers.local_worker().env
                epp: EpisodeParameterProvider = env.config.epp
                epp.load_checkpoint(checkpoint_folder / epp_name)

                # Agents
                for agent_name, agent_configs in env.agent_dict.items():
                    epp = agent_configs.config.epp
                    epp.load_checkpoint(checkpoint_folder / agent_name)

        self.config.tune_config["run_or_experiment"] = EpisodeParameterProviderSavingTrainer

    def _add_trial_creator(self):
        """Updates the trial name based on the HPC Job Number and the trial name in the configuration
        """
        if "trial_name_creator" not in self.config.tune_config:
            if self.config.trial_creator_function is None:

                def trial_name_prefix(trial):
                    """
                    Args:
                        trial (Trial): A generated trial object.

                    Returns:
                        trial_name (str): String representation of Trial prefixed
                            by the contents of the environment variable:
                            TRIAL_NAME_PREFIX
                            Or the prefix 'RUN' if none is set.
                    """
                    trial_prefix = os.environ.get('PBS_JOBID', os.environ.get('TRIAL_NAME_PREFIX', ""))
                    trial_name = ""

                    if "TrialName" in self.config.env_config.keys():
                        if trial_prefix:
                            trial_name = "-" + self.config.env_config["TrialName"]
                        else:
                            trial_name = self.config.env_config["TrialName"]
                    return f"{trial_prefix}{trial_name}-{trial}"

                self.config.tune_config["trial_name_creator"] = trial_name_prefix
            else:
                self.config.tune_config["trial_name_creator"] = self.config.trial_creator_function

    def _add_experiment_output_dir(self, config: SafeAutonomyRllibExperimentValidator):
        now = datetime.now()

        expr_dir_name = f"expr_{now.strftime('%Y%m%d_%H%M%S')}"

        config.env_config['output_path'] = os.path.join(config.env_config['output_path'], expr_dir_name)
        config.tune_config['local_dir'] = os.path.join(config.tune_config['local_dir'], expr_dir_name)
        return config



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains the CWH Simulator for interacting with the CWH Docking task simulator
"""

from corl.libraries.plugin_library import PluginLibrary
from safe_autonomy_simulation.sims.spacecraft import CWHSpacecraft

from safe_autonomy_sims.platforms.cwh.cwh_platform import CWHPlatform
from safe_autonomy_sims.simulators.saferl_simulator import SafeRLSimulator


class CWHSimulator(SafeRLSimulator):
    """
    Simulator for CWH Docking Task. Interfaces CWH platforms with underlying CWH entities in Docking simulation.
    """

    def _construct_platform_map(self) -> dict:
        return {
            'default': (CWHSpacecraft, CWHPlatform),
            'cwh': (CWHSpacecraft, CWHPlatform),
        }


PluginLibrary.AddClassToGroup(CWHSimulator, "CWHSimulator", {})



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements Runtime Assurance modules used in
the docking and inspection environments.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements Run Time Assurance for the inspection task.
"""
import math
from typing import Union

import numpy as np
from run_time_assurance.utils import to_jnp_array_jit
from run_time_assurance.zoo.cwh.inspection_1v1 import (
    CHIEF_RADIUS_DEFAULT,
    DEPUTY_RADIUS_DEFAULT,
    FOV_DEFAULT,
    M_DEFAULT,
    N_DEFAULT,
    R_MAX_DEFAULT,
    SUN_VEL_DEFAULT,
    U_MAX_DEFAULT,
    V0_DEFAULT,
    V0_DISTANCE_DEFAULT,
    V1_COEF_DEFAULT,
    VEL_LIMIT_DEFAULT,
    Inspection1v1RTA,
)

from safe_autonomy_sims.glues.rta_glue import RTAGlue, RTAGlueValidator, RTASingleton


class CWHInspection1v1RTAGlueValidator(RTAGlueValidator):
    """
    A configuration validator for CWHInspection1v1RTAGlue.

    Attributes
    ----------
    m : float, optional
        mass in kg of spacecraft, by default M_DEFAULT
    n : float, optional
        orbital mean motion in rad/s of current Hill's reference frame, by default N_DEFAULT
    chief_radius : float, optional
        radius of collision for chief spacecraft, by default CHIEF_RADIUS_DEFAULT
    deputy_radius : float, optional
        radius of collision for each deputy spacecraft, by default DEPUTY_RADIUS_DEFAULT
    collision_radius : Union[float, None], optional
        Total collision radius - alternative to defining chief radius and deputy radius separately. By default None
    v0 : float, optional
        Maximum safe docking velocity in m/s, by default V0_DEFAULT
        v0 of v_limit = v0 + v1*n*||r-v0_distance||
    v1_coef : float, optional
        coefficient of linear component of the distance depending speed limit in 1/seconds, by default V1_COEF_DEFAULT
        v1_coef of v_limit = v0 + v1_coef*n*||r-v0_distance||
    v0_distance: float
        NMT safety constraint minimum distance where v0 is applied. By default 0.
    r_max : float, optional
        maximum relative distance from chief, by default R_MAX_DEFAULT
    fov : float, optional
        sensor field of view, by default FOV_DEFAULT
    vel_limit : float, optional
        max velocity magnitude, by default VEL_LIMIT_DEFAULT
    sun_vel : float, optional
        velocity of sun in x-y plane (rad/sec), by default SUN_VEL_DEFAULT
    control_bounds_high : float, optional
        upper bound of allowable control. Pass a list for element specific limit. By default U_MAX_DEFAULT
    control_bounds_low : float, optional
        lower bound of allowable control. Pass a list for element specific limit. By default -U_MAX_DEFAULT
    constraints : list, optional
        list of constraint keys to include
    """
    m: float = M_DEFAULT
    n: float = N_DEFAULT
    chief_radius: float = CHIEF_RADIUS_DEFAULT
    deputy_radius: float = DEPUTY_RADIUS_DEFAULT
    collision_radius: Union[float, None] = None
    v0: float = V0_DEFAULT
    v0_distance: float = V0_DISTANCE_DEFAULT
    v1_coef: float = V1_COEF_DEFAULT
    r_max: float = R_MAX_DEFAULT
    fov: float = FOV_DEFAULT
    vel_limit: float = VEL_LIMIT_DEFAULT
    sun_vel: float = SUN_VEL_DEFAULT
    control_bounds_high: float = U_MAX_DEFAULT
    control_bounds_low: float = -U_MAX_DEFAULT
    constraints: list = []


class RTAGlueCWHInspection1v1(RTAGlue):
    """
    RTA Glue to wrap CWH Inspection 1v1 RTA from the run-time-assurance package.
    """

    def __init__(self, **kwargs):
        self.config: CWHInspection1v1RTAGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return CWHInspection1v1RTAGlueValidator

    def _get_singleton(self):
        return InspectionRTASingleton

    def _get_rta_args(self) -> dict:
        if self.config.collision_radius is not None:
            self.config.chief_radius = self.config.collision_radius / 2
            self.config.deputy_radius = self.config.collision_radius / 2

        return {
            "m": self.config.m,
            "n": self.config.n,
            "chief_radius": self.config.chief_radius,
            "deputy_radius": self.config.deputy_radius,
            "v0": self.config.v0,
            "v0_distance": self.config.v0_distance,
            "v1_coef": self.config.v1_coef,
            "r_max": self.config.r_max,
            "fov": self.config.fov,
            "vel_limit": self.config.vel_limit,
            "sun_vel": self.config.sun_vel,
            "control_bounds_high": self.config.control_bounds_high,
            "control_bounds_low": self.config.control_bounds_low,
            "constraints_to_use": self.config.constraints,
        }


class InspectionRTASingleton(RTASingleton):
    """Inspection Singleton
    """

    def _create_rta_module(self, **rta_args):
        return UpdatedInspection1v1RTA(**rta_args)


class UpdatedInspection1v1RTA(Inspection1v1RTA):
    """Updated RTA module with _get_state"""

    def _get_state(self, input_state):
        if len(input_state) == 7:
            input_state = np.array(input_state)
            input_state = np.concatenate((input_state, np.array([0.])))
            theta = input_state[6]
            m = math.floor(abs(theta / (2 * np.pi)))
            theta -= np.sign(theta) * m * 2 * np.pi
            input_state[6] = 2 * np.pi - theta
        else:
            input_state = np.concatenate((input_state, np.array([0., 0.])))
        return to_jnp_array_jit(input_state)



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module defines the platforms used with safe_autonomy_sims CWHSimulator class.
It represents a spacecraft operating under the Clohessy-Wiltshire dynamics model, with and without rotation.
"""

import typing

import numpy as np
from corl.libraries.units import corl_get_ureg
from corl.simulators.base_platform import BasePlatformValidator
from safe_autonomy_simulation.sims.spacecraft import CWHSpacecraft, SixDOFSpacecraft

from safe_autonomy_sims.platforms.common.platform import BaseSafeRLPlatform


class CWHPlatformValidator(BasePlatformValidator):
    """
    A configuration validator for CWHPlatform

    Attributes
    ----------
    platform : CWHSpacecraft
        underlying dynamics platform
    """

    platform: CWHSpacecraft


class CWHPlatform(BaseSafeRLPlatform):
    """
    A platform representing a spacecraft operating under CWH dynamics.
    Allows for saving an action to the platform for when the platform needs
    to give an action to the environment during the environment step function.
    """

    def __init__(self, platform_name, platform, parts_list, sim_time=0.0):
        """
        Parameters
        ----------
        platform_name : str
            Name of the platform.
        platform : sim_entity
            Backend simulation entity associated with the platform.
        parts_list : dict
            Platform-specific parts configuration dictionary.
        sim_time : float
            simulation time at platform creation
        """
        self.config: CWHPlatformValidator
        super().__init__(platform_name=platform_name, platform=platform, parts_list=parts_list, sim_time=sim_time)
        self._platform = self.config.platform
        self._last_applied_action = corl_get_ureg().Quantity(np.array([0, 0, 0], dtype=np.float32), "newtons")

    @staticmethod
    def get_validator() -> typing.Type[CWHPlatformValidator]:
        """
        get validator for this CWHPlatform

        Returns
        -------
        CWHPlatformValidator -- validator the platform will use to generate a configuration
        """
        return CWHPlatformValidator

    def get_applied_action(self):
        """
        Returns the action stored in this platform.

        Returns
        -------
        typing.Any -- Any sort of stored action.
        """
        return self._last_applied_action

    def __eq__(self, other):
        if isinstance(other, CWHPlatform):
            eq = (self.velocity == other.velocity).all()
            eq = eq and (self.position == other.position).all()
            eq = eq and self.sim_time == other.sim_time
            return eq
        return False

    def save_action_to_platform(self, action, axis):
        """
        Saves an action to the platform if it matches the action space.

        Parameters
        ----------
        action: typing.Any
            The action to store in the platform
        axis: int
            The index of the action space where the action shall be saved.
        """
        if isinstance(action.m, np.ndarray) and len(action.m) == 1:
            # TODO: is there a way to ensure quantity coming in is of correct units?
            #       currently, incoming action is 'dimensionless'.
            # action = action.to('newton')
            self._last_applied_action.m[axis] = action.m[0]
        else:
            raise TypeError(
                f"Action saved to platform is of incompatible type:\
                             expected Quantity with numpy.ndarray of length 1, but got {type(action.m)}"
            )

    @property
    def position(self) -> np.ndarray:
        """
        The position of the platform.

        Returns
        -------
        np.ndarray
            The position vector of the platform.
        """
        return self._platform.position

    @property
    def velocity(self):
        """
        The velocity of the platform.

        Returns
        -------
        np.ndarray
            The velocity vector of the platform.
        """
        return self._platform.velocity

    @property
    def operable(self):
        return True

    def entity_relative_position(self, entity_name):
        """
        The position of entity_name relative to self (without rotation)

        Returns
        -------
        np.ndarray
            The relative position of entity_name
        """
        return self._platform.entity_relative_position(entity_name)

    def entity_relative_velocity(self, entity_name):
        """
        The velocity of entity_name relative to self (without rotation)

        Returns
        -------
        np.ndarray
            The relative velocity of entity_name
        """
        return self._platform.entity_relative_velocity(entity_name)


class CWHSixDOFPlatformValidator(BasePlatformValidator):
    """
    A configuration validator for CWHSixDOFPlatform

    Attributes
    ----------
    platform : SixDOFSpacecraft
        underlying dynamics platform
    """
    platform: SixDOFSpacecraft


class CWHSixDOFPlatform(CWHPlatform):
    """
    A platform representing a spacecraft operating under CWH dynamics with rotation.
    Allows for saving an action to the platform for when the platform needs
    to give an action to the environment during the environment step function.
    """

    def __init__(self, platform_name, platform, parts_list, sim_time=0.0):
        """
        Parameters
        ----------
        platform_name : str
            Name of the platform.
        platform : sim_entity
            Backend simulation entity associated with the platform.
        parts_list : dict
            Platform-specific parts configuration dictionary.
        sim_time : float
            simulation time at platform creation
        """
        self.config: CWHSixDOFPlatformValidator
        super().__init__(platform_name=platform_name, platform=platform, parts_list=parts_list, sim_time=sim_time)
        self._last_applied_action = corl_get_ureg().Quantity(np.array([0, 0, 0, 0, 0, 0], dtype=np.float32), "newtons")

    @staticmethod
    def get_validator() -> typing.Type[CWHSixDOFPlatformValidator]:
        """
        get validator for this CWHPlatform

        Returns:
            CWHPlatformValidator -- validator the platform will use to generate a configuration
        """
        return CWHSixDOFPlatformValidator

    def __eq__(self, other):
        if isinstance(other, CWHSixDOFPlatform):
            eq = (self.velocity == other.velocity).all()
            eq = eq and (self.position == other.position).all()
            eq = eq and self.sim_time == other.sim_time
            return eq
        return False

    @property
    def quaternion(self):
        """
        The quaternion of the platform.

        Returns
        -------
        np.ndarray
            The quaternion vector of the platform.
        """
        return self._platform.quaternion

    @property
    def angular_velocity(self):
        """
        The angular_velocity of the platform.

        Returns
        -------
        np.ndarray
            The angular_velocity vector of the platform.
        """
        return self._platform.angular_velocity



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains a parameter wrapper for RTA rejection sampling.
If chosen initial parameters violate safety, they are rejected and resampled.
"""

import typing
from collections import OrderedDict

import numpy as np
from corl.libraries.parameters import ParameterWrapper, ParameterWrapperValidator
from corl.libraries.units import Quantity
from numpy.random import Generator, RandomState
from pydantic import BaseModel, ConfigDict, create_model
from pydantic.types import PyObject
from run_time_assurance.rta import ConstraintBasedRTA
from run_time_assurance.utils import to_jnp_array_jit

from safe_autonomy_sims.simulators.initializers.initializer import BaseInitializer, InitializerValidator
from safe_autonomy_sims.simulators.saferl_simulator import InitializerResetValidator

Randomness = Generator | RandomState
OtherVars = typing.Mapping[tuple[str, ...], Quantity]


class RTAValidator(BaseModel):
    """
    A configuration validator for RTA glue object

    Attributes
    ----------
    functor: PyObject
        The class module of the RTA Glue to be instantiated
    states: list
        States to be passed from the initializer to the RTA module (in order)
    args: dict
        Arguments used to initialize the RTA module
    arg_map: dict
        Mapping of arguments from 'args' to kwargs accepted by the RTA module.
        Useful when 'args' contains extra values or naming is different.
        By default None.
    """
    functor: PyObject
    states: list
    args: dict
    arg_map: typing.Union[dict, None] = None


class RejectionSamplerWrapper(ParameterWrapperValidator):
    """
    A configuration validator for rejection sampler parameter wrapper

    Attributes
    ----------
    initializer: InitializerResetValidator
        Initializer to convert parameters to values used by the RTA
    rta: RTAValidator
        RTA glue used to check safety constraint values
    max_rejections: int
        Maximum number of times to reject a sample before passing it through
    """
    initializer: InitializerResetValidator
    rta: RTAValidator
    max_rejections: int = 1000


class RejectionSampler(ParameterWrapper):
    """
    Rejection Sampling Parameter Wrapper.
    If sampled parameters violate safety constraints, they are rejected and resampled.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # Reorganize args if necessary
        if self.config.rta.arg_map is not None:
            new_args = {}
            for k, v in self.config.rta.arg_map.items():
                new_args[k] = self.config.rta.args[v]
            self.config.rta.args = new_args

        # Set unused args for RTA glue
        self.config.rta.args['wrapped'] = []
        self.config.rta.args['agent_name'] = 'None'
        self.config.rta.args['state_observation_names'] = ['None']

    @staticmethod
    def get_validator() -> type[ParameterWrapperValidator]:
        """Get the validator class for this Parameter"""
        return RejectionSamplerWrapper

    def get_value(self, rng: Randomness, other_vars: OtherVars) -> dict[str, str | Quantity]:
        # Create initializer (errors when this is done in __init__)
        initializer = self.config.initializer.functor(self.config.initializer.config)

        safe_state = False
        check_counter = 0
        # Loop until safe state is found or max iterations is reached
        while not safe_state and check_counter < self.config.max_rejections:
            check_counter += 1
            sample = self.get_sample_value(rng, other_vars, initializer)
            state = self.get_state(sample)
            safe_state = self.check_if_safe_state(state)

        # TODO: fixes callback error
        sample['m'] = 0.0

        # We create a dynamic model since the rejection sampler can wrap any parameter as defined
        # in configuration files
        field_defs = OrderedDict({k: (type(v), v) for k, v in sample.items()})
        model_config = ConfigDict(arbitrary_types_allowed=True)
        params_model = create_model('init_state', **field_defs, __config__=model_config)(**sample)

        return params_model  # type: ignore

    def get_sample_value(self, rng: Randomness, other_vars: OtherVars, initializer: BaseInitializer) -> typing.Dict:
        """
        Sample the parameters

        Parameters
        ----------
        rng: Randomness
            Randomness for sampling
        other_vars: OtherVars
            Other variables for sampled parameters
        initializer: BaseInitializer
            Initializer used to get state from parameters

        Returns
        -------
        typing.Dict
            Initializer state
        """
        params = {}
        for k, p in self.config.wrapped.items():
            v = p.get_value(rng, other_vars)
            params[k] = v

        transformed_params = initializer.compute(**params)
        return transformed_params

    def get_state(self, value: typing.Dict) -> np.ndarray:
        """
        Gets RTA state from Initializer state

        Parameters
        ----------
        value: typing.Dict
            Initializer state

        Returns
        -------
        np.ndarray
            RTA state
        """
        array = []
        for k in self.config.rta.states:
            array.append(value[k])
        return np.array(array)

    def check_if_safe_state(self, state: np.ndarray):
        """
        Determines if state is safe or not.

        Parameters
        ----------
        state: np.ndarray
            Current system state

        Returns
        -------
        bool
            True if state is safe, False if not.
        """
        assert isinstance(self.config.rta.functor(**self.config.rta.args).rta,
                          ConstraintBasedRTA), ("Must use constraint based rta for rejection sampling.")
        init_state_safe = True

        # For each constraint, check if satisfied
        for c in self.config.rta.functor(**self.config.rta.args).rta.constraints.values():
            if c.phi(to_jnp_array_jit(state), c.params) < 0 or c(to_jnp_array_jit(state), c.params) < 0:
                init_state_safe = False
                break
        return init_state_safe


class RejectionSamplerInitializerValidator(InitializerValidator):
    """
    A configuration validator for agent initializer to be used with the rejection sampler

    Attributes
    ----------
    states: typing.Optional[typing.List]
        State keys to initialize the agent.
        Useful when the rejection sampler state contains extra values.
        By default None.
    """
    states: typing.Optional[typing.List] = None


class RejectionSamplerInitializer(BaseInitializer):
    """
    Agent initializer to pass through init_state from rejection sampler.
    Example config usage:

    "simulator_reset_parameters": {
      "initializer": {
        "functor": "safe_autonomy_sims.rta.rta_rejection_sampler.RejectionSamplerInitializer",
        "config": {}
      },
      "config": {
        "init_state": {
          "type": "safe_autonomy_sims.simulators.initializers.initializer.SimAttributeAccessor",
          "config": {
            "attribute_name": "init_state",
          }
        }
      }
    }
    """

    @staticmethod
    def get_validator() -> typing.Type[InitializerValidator]:
        return RejectionSamplerInitializerValidator

    def compute(self, **kwargs):
        # Convert to dict
        init_state = dict(kwargs['init_state'])
        if self.config.states is not None:
            states = {}
            for k in self.config.states:
                states[k] = init_state[k]
        else:
            states = init_state

        return states



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a glue for filtering an agent's action through a Run Time Assurance filter.
"""
import abc
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np
from corl.glues.base_multi_wrapper import BaseMultiWrapperGlue, BaseMultiWrapperGlueValidator
from corl.glues.common.controller_glue import ControllerGlue
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import Prop
from corl.libraries.units import corl_get_ureg
from run_time_assurance.rta import ConstraintBasedRTA, RTAModule

from safe_autonomy_sims.platforms.cwh.cwh_properties import TupleProp


def flip_rta(control):
    """
    Simple filter which flips the sign of the input control.

    Parameters
    ----------
    control: dict
        Dictionary of input controls.

    Returns
    -------
    dict
        The filtered control dictionary.
    """
    filtered_control = OrderedDict()
    for k, v in control.items():
        filtered_control[k] = -v
    return filtered_control


class RTAGlueValidator(BaseMultiWrapperGlueValidator):
    """
    A configuration validator for RTAGlue.

    Attributes
    ----------
    step_size: float
        duration in seconds that agent's action will be applied
    state_observation_names: list[str]
        list of keys from observation dict whose direct observation values will be concatenated to form the RTA state vector
    enabled: bool
        True if RTA is enabled
    """
    step_size: float
    state_observation_names: typing.List[str]
    enabled: bool = True


class RTAGlue(BaseMultiWrapperGlue):
    """
    Glue containing RTA module for filtering actions.
    """

    def __init__(self, **kwargs):
        self.config: RTAGlueValidator
        super().__init__(**kwargs)
        self.controller_glues = self._get_controller_glues(self)
        rta_args = self._get_rta_args()
        singleton = self._get_singleton()
        self.rta = singleton.instance(**rta_args).rta
        self.rta.enable = self.config.enabled
        self.filtered_action = None

    @staticmethod
    def get_validator():
        return RTAGlueValidator

    def get_unique_name(self) -> str:
        return "RTAModule"

    @cached_property
    def action_prop(self) -> typing.Optional[Prop]:
        """
        Build the action property for the controllers that defines the action this glue produces

        Returns
        -------
        typing.Optional[Prop]
            The Property that defines what this glue requires for an action
        """
        # collect action properties of wrapped controllers
        action_props = tuple(glue.action_prop for glue in self.glues())
        prop = TupleProp(spaces=action_props)
        return prop

    @cached_property
    def action_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        action_spaces = [gymnasium.spaces.Dict({glue.action_prop.name: glue.action_space}) for glue in self.glues()]
        return gymnasium.spaces.tuple.Tuple(tuple(action_spaces))

    def controller_glue_action_space(self) -> gymnasium.spaces.tuple.Tuple:
        """
        Compiles the action spaces for the terminal control glues for each wrapped chain of glues
        i.e. compiles the action spaces that directly interface with the platform actuators

        Returns
        -------
        gymnasium.spaces.Space
            The gymnasium Space that defines the actions given to the apply_action function for the wrapped terminal controller glues
        """
        action_spaces = [
            gymnasium.spaces.Dict({controller_glue.action_prop.name: controller_glue.action_space})
            for controller_glue in self.controller_glues
        ]
        return gymnasium.spaces.tuple.Tuple(tuple(action_spaces))

    def apply_action(
        self,
        action: EnvSpaceUtil.sample_type,
        observation: EnvSpaceUtil.sample_type,
        action_space: OrderedDict,
        obs_space: OrderedDict,
        obs_units: OrderedDict
    ) -> None:
        assert isinstance(action, tuple)

        for i in range(len(self.glues())):
            glue = self.glues()[i]
            glue.apply_action(action[i][glue.action_prop.name], observation, action_space, obs_space, obs_units)

        desired_action = self._get_stored_action()
        filtered_action = self._filter_action(desired_action, observation)

        for controller_glue, controller_filtered_action in zip(self.controller_glues, filtered_action):
            controller_glue.apply_action(
                controller_filtered_action[controller_glue.action_prop.name], observation, action_space, obs_space, obs_units
            )

    def _filter_action(self, desired_action: tuple, observation: typing.Dict) -> tuple:
        rta_state_vector = self._get_rta_state_vector(observation)
        rta_action_vector = self._get_action_vector_from_action(desired_action).flatten()
        filtered_action_vector = self.rta.filter_control(rta_state_vector, self.config.step_size, rta_action_vector)
        if isinstance(self.rta, ConstraintBasedRTA):
            self.rta.update_constraint_values(rta_state_vector)
        return self._get_action_from_action_vector(filtered_action_vector)

    def _get_rta_state_vector(self, observation: typing.Dict) -> np.ndarray:
        state_obs = []
        for obs_name in self.config.state_observation_names:
            try:
                state_obs.append(observation[obs_name]['direct_observation'].m)
            except KeyError as e:
                raise KeyError(f"state observation {obs_name} not found. Must be one of {list(observation.keys())}") from e

        state_vec = np.concatenate(state_obs)
        return state_vec

    def _get_action_vector_from_action(self, action: tuple) -> np.ndarray:
        actions_ordered = []
        for controller_action in action:
            actions_ordered.append(controller_action.m)
        control_vector = np.array(actions_ordered)
        return control_vector

    def _get_action_from_action_vector(self, combined_action_vector: np.ndarray) -> tuple:
        combined_action_left = combined_action_vector
        controller_action_spaces = self.controller_glue_action_space()
        action_list = []

        for controller_action_space in controller_action_spaces:
            controller_action = OrderedDict()
            for action_key, action_space in controller_action_space.items():
                action_length = np.prod(action_space.shape)
                action_value = combined_action_left[:action_length]
                combined_action_left = combined_action_left[action_length:]
                controller_action[action_key] = corl_get_ureg().Quantity(action_value, "dimensionless")
            action_list.append(controller_action)
        return tuple(action_list)

    @abc.abstractmethod
    def _get_singleton(self):
        """
        Get singleton class containing RTA module.

        Returns
        -------
        RTASingleton
            Custom singleton.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _get_rta_args(self) -> dict:
        """
        Get RTA arguments.

        Returns
        -------
        dict
            RTA args to be passed at initialization.
        """
        raise NotImplementedError

    @cached_property
    def normalized_action_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        return self.action_space

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        # TODO: Add info to callback for eval metrics
        # obs = {"intervening": int(self.rta.intervening)}
        # if isinstance(self.rta, ConstraintBasedRTA):
        #     info = self.rta.generate_info()['constraints'].items()
        #     n_val = np.array([-1.], dtype=np.float32)
        #     obs['constraints'] = {k: np.array([v], dtype=np.float32) if not np.isnan(v) else n_val for k, v in info}
        # return obs
        return None

    def get_info_dict(self):
        return {
            "actual control": self.rta.control_actual,
            "desired control": self.rta.control_desired,
            "intervening": self.rta.intervening,
        }

    def _get_controller_glues(self, glue):
        controller_glues = []
        if isinstance(glue, ControllerGlue):
            controller_glues.append(glue)
        else:
            wrapped_list = list(glue.config.wrapped)
            for wrapped_glue in wrapped_list:
                controller_glues.extend(self._get_controller_glues(glue=wrapped_glue))
        return controller_glues

    def _get_stored_action(self) -> tuple:
        stored_action = []
        for controller_glue in self.controller_glues:
            applied_action = controller_glue.get_applied_control()
            stored_action.append(applied_action)
        return tuple(stored_action)


class RTASingleton(abc.ABC):
    """
    Implement RTA module as a singleton.
    This prevents reinitialization, which can cause memory leaks due to jit.
    """

    _instance = None

    @classmethod
    def instance(cls, **rta_args):
        """Create singleton"""
        if not cls._instance:
            cls._instance = cls.__new__(cls)
            cls.rta = cls._create_rta_module(cls._instance.__class__, **rta_args)
        return cls._instance

    def __init__(self, **kwargs):
        raise RuntimeError("Use the 'instance' method to initialize RTA objects.")

    @abc.abstractmethod
    def _create_rta_module(self, **rta_args: dict) -> RTAModule:
        """
        Initialize the RTA module

        Parameters
        ----------
        rta_args: dict
            Dictionary of args passed to RTA.

        Returns
        -------
        RTAModule
            Custom RTA module.
        """
        raise NotImplementedError
